/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
* Version: NPL 1.1/GPL 2.0/LGPL 2.1
*
* The contents of this file are subject to the Netscape Public License
* Version 1.1 (the "License"); you may not use this file except in
* compliance with the License. You may obtain a copy of the License at
* http://www.mozilla.org/NPL/
*
* Software distributed under the License is distributed on an "AS IS" basis,
* WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
* for the specific language governing rights and limitations under the
* License.
*
* The Original Code is mozilla.org code.
*
* The Initial Developer of the Original Code is 
* Netscape Communications Corporation.
* Portions created by the Initial Developer are Copyright (C) 2002
* the Initial Developer. All Rights Reserved.
*
* Contributor(s):
*   Geoff Beier <me@mollyandgeoff.com>
*
* Alternatively, the contents of this file may be used under the terms of
* either the GNU General Public License Version 2 or later (the "GPL"), or 
* the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
* in which case the provisions of the GPL or the LGPL are applicable instead
* of those above. If you wish to allow use of your version of this file only
* under the terms of either the GPL or the LGPL, and not to allow others to
* use your version of this file under the terms of the NPL, indicate your
* decision by deleting the provisions above and replace them with the notice
* and other provisions required by the GPL or the LGPL. If you do not delete
* the provisions above, a recipient may use your version of this file under
* the terms of any one of the NPL, the GPL or the LGPL.
*
* ***** END LICENSE BLOCK ***** */

#import "BrowserTabBarView.h"
#import "BrowserTabViewItem.h"
#import "TabButtonCell.h"


@interface BrowserTabBarView (PRIVATE)
-(void) layoutButtons;
-(void) loadImages;
-(void) drawTabBarBackgroundInRect:(NSRect)rect withActiveTabRect:(NSRect)tabRect;
-(TabButtonCell *) buttonAtPoint:(NSPoint)clickPoint;
-(void) registerTabButtonsForTracking;
-(void) unregisterTabButtonsForTracking;
@end

static NSImage *backgroundImage = nil;
static NSImage *tabButtonDividerImage = nil;
static float tabBarDefaultHeight = 22.0;

@implementation BrowserTabBarView

#define TAB_BAR_MARGIN 5    // left/right margin for tab bar
#define TAB_GAP 2           // space between tabs
#define MIN_TAB_WIDTH 16    // tabs smaller than this are useless... actually, tabs this small are probably useless :-)
#define MAX_TAB_WIDTH 256   // the widest tabs that will be drawn

- (id)initWithFrame:(NSRect)frame 
{
    self = [super initWithFrame:frame];
    if (self) {
      mTabButtons = nil;
      mActiveTabButton = nil;
      // this will not likely have any result here
      [self rebuildTabBar];
    }
    return self;
}

- (void)awakeFromNib
{
  // this needs to be called again since our tabview should be non-nil now
  [self rebuildTabBar];
}

- (void)dealloc
{
  [mTabButtons release];
  [mActiveTabButton release];
  [super dealloc];
}

- (void)drawRect:(NSRect)rect 
{
  // this should only occur the first time any instance of this view is drawn
  if( (nil == backgroundImage)||(nil == tabButtonDividerImage) ){
    [self loadImages];
  }
  
  // determine the frame of the active tab button and fill the rest of the bar in with the background
  NSRect activeTabButtonFrame = [mActiveTabButton frame];  
  [self drawTabBarBackgroundInRect:rect withActiveTabRect:activeTabButtonFrame];
  
  NSEnumerator *tabEnumerator = [mTabButtons objectEnumerator];
  TabButtonCell *tabButton = [tabEnumerator nextObject];
  while( nil != tabButton ) {
    
    TabButtonCell *nextTab = [tabEnumerator nextObject];
    
    NSRect tabButtonFrame = [tabButton frame];
    if( NSIntersectsRect(tabButtonFrame,rect) ) {
      [tabButton drawWithFrame:tabButtonFrame inView:self];
    }
    
    // draw the divider if we're not the selected tab... the selected tab handles its own
    if( tabButton != mActiveTabButton){
      [tabButtonDividerImage compositeToPoint:NSMakePoint( tabButtonFrame.origin.x, tabButtonFrame.origin.y )
                                    operation:NSCompositeSourceOver];
    }
    if( !nextTab ) {
      [tabButtonDividerImage compositeToPoint:NSMakePoint( NSMaxX(tabButtonFrame), tabButtonFrame.origin.y )
                                    operation:NSCompositeSourceOver];
    }    
    tabButton = nextTab;
  }
}

- (void)setFrame:(NSRect)frameRect
{
  [super setFrame:frameRect];
  // tab buttons probably need to be resized if the frame changes
  [self layoutButtons];
}

- (NSMenu *)menuForEvent:(NSEvent *)theEvent
{
  NSPoint clickPoint = [self convertPoint:[theEvent locationInWindow] fromView:nil];
  TabButtonCell * clickedTabButton = [self buttonAtPoint:clickPoint];
  if(clickedTabButton) {
    return [clickedTabButton menu];
  }
  // XXX should there be a menu if someone clicks in the tab bar where there is no tab?
  return nil;
}


- (void)mouseDown:(NSEvent *)theEvent
{
  NSPoint clickPoint = [self convertPoint:[theEvent locationInWindow] fromView:nil];
  TabButtonCell * clickedTabButton = [self buttonAtPoint:clickPoint];
  //TODO: start drag tracking
  if(clickedTabButton) {
    if(![clickedTabButton willTrackMouse:theEvent inRect:[clickedTabButton frame] ofView:self]){
      [mTabView selectTabViewItem:[clickedTabButton tabViewItem]];
    }
  }
}

// returns the tab at the specified point
- (TabButtonCell *) buttonAtPoint:(NSPoint)clickPoint
{
  NSEnumerator * buttonEnumerator;
  TabButtonCell * button;
  buttonEnumerator = [mTabButtons objectEnumerator];
  while( button = [buttonEnumerator nextObject] ) {
    if( NSPointInRect(clickPoint,[button frame] ) ){
      return button;
    }
  }
  return nil;
}

- (void) drawTabBarBackgroundInRect:(NSRect)rect withActiveTabRect:(NSRect)tabRect
{
  // draw tab bar background, omitting the selected Tab
  NSRect barFrame = [self bounds];
  NSRect fillRect;
  
  // first, fill to the left of the active tab
  fillRect = NSMakeRect( barFrame.origin.x, barFrame.origin.y, 
                         (tabRect.origin.x - barFrame.origin.x), barFrame.size.height);
  if( NSIntersectsRect(fillRect,rect) ) {
    // make sure we're not drawing to the left or right of the actual rectangle we were asked to draw
    if( fillRect.origin.x < NSMinX(rect) ) {
      fillRect.size.width -= NSMinX(rect) - fillRect.origin.x;
      fillRect.origin.x = NSMinX(rect);
    }
    if( NSMaxX(fillRect) > NSMaxX(rect) ) {
      fillRect.size.width -= NSMaxX(fillRect) - NSMaxX(rect);
    }
    [backgroundImage paintTiledInRect:fillRect];
  }
  // then fill to the right
  fillRect = NSMakeRect( NSMaxX(tabRect), barFrame.origin.y, 
                           (NSMaxX(barFrame) - NSMaxX(tabRect)), barFrame.size.height);
  if( NSIntersectsRect(fillRect,rect) ) {
      // make sure we're not drawing to the left or right of the actual rectangle we were asked to draw
      if( fillRect.origin.x < NSMinX(rect) ) {
        fillRect.size.width -= NSMinX(rect) - fillRect.origin.x;
        fillRect.origin.x = NSMinX(rect);
      }
      if( NSMaxX(fillRect) > NSMaxX(rect) ) {
        fillRect.size.width -= NSMaxX(fillRect) - NSMaxX(rect);
      }
      [backgroundImage paintTiledInRect:fillRect];
   }
}
  
      

- (void)loadImages
{
  backgroundImage = [[NSImage imageNamed:@"tab_bar_bg"] retain];
  tabButtonDividerImage = [[NSImage imageNamed:@"tab_button_divider"] retain];
}

// construct the tab bar based on the current state of mTabView;
// should be called when tabs are first shown.
// returns YES if it was possible to rebuild the toolbar
- (void)rebuildTabBar
{
  [self unregisterTabButtonsForTracking];
  [mTabButtons release];
  [mActiveTabButton release];
  
  mTabButtons = [[NSMutableArray alloc] init];
  
  for( int i = 0; i < [mTabView numberOfTabViewItems]; i++ ) {
    BrowserTabViewItem *page = [mTabView tabViewItemAtIndex:i];
    // autorelease so that deallocing the array actually calls the dealloc method on all the items released
    TabButtonCell *tabButton = [[[TabButtonCell alloc] initFromTabViewItem:page] autorelease];
    // keep a reference to it if it's the currently selected tab
    if( [tabButton isSelected] ) {
      mActiveTabButton = [tabButton retain];
    }
    [tabButton setFrame:NSMakeRect(0,0,MAX_TAB_WIDTH,tabBarDefaultHeight)];
    [mTabButtons addObject:tabButton];
  }
  [self layoutButtons];
  [self registerTabButtonsForTracking];
}

// allows tab button cells to react to mouse events
- (void) registerTabButtonsForTracking
{
  if( [self window] ) {
    NSEnumerator * tabButtonEnumerator = [mTabButtons objectEnumerator];
    
    NSPoint local = [[self window] convertScreenToBase:[NSEvent mouseLocation]];
    local = [self convertPoint:local fromView:nil];
    
    TabButtonCell *tabButton;
    while( tabButton = [tabButtonEnumerator nextObject] ) {
      NSRect trackingRect = [tabButton frame];
      [tabButton addTrackingRectInView: self withFrame:trackingRect cursorLocation:local];
    }
  }
}

// causes tab buttons to stop reacting to mouse events
- (void) unregisterTabButtonsForTracking
{
  NSEnumerator * tabButtonEnumerator = [mTabButtons objectEnumerator];
  TabButtonCell *tabButton;
  while( tabButton = [tabButtonEnumerator nextObject] ) {
    [tabButton removeTrackingRectFromView: self];
  }
}
  
// returns the height the tab bar should be if drawn
- (float)tabBarHeight
{
  // this will be constant for now
  return tabBarDefaultHeight;
}

// sets the tab buttons to the largest MIN_TAB_WIDTH <= size <= MAX_TAB_WIDTH where they all fit
// and calculates the frame for each.
- (void)layoutButtons
{
  int numTabs = [mTabView numberOfTabViewItems];
  float tabWidth = MAX_TAB_WIDTH;

  // calculate the largest tabs that would fit
  float maxWidth = ((NSWidth([self bounds]) - (2*TAB_BAR_MARGIN) - ((numTabs - 1)*TAB_GAP))/numTabs);
  // if our tabs are currently larger than that, shrink them to the larger of MIN_TAB_WIDTH or maxWidth
  if( tabWidth > maxWidth ) {
    tabWidth = (maxWidth > MIN_TAB_WIDTH ? maxWidth : MIN_TAB_WIDTH);
  }
  
  // resize and position the tab buttons
  int xCoord = TAB_BAR_MARGIN;
  NSEnumerator *tabEnumerator = [mTabButtons objectEnumerator];
  TabButtonCell *tabButtonCell;
  
  while( (tabButtonCell = [tabEnumerator nextObject]) ) {
    NSSize buttonSize = [tabButtonCell size];
    buttonSize.width = tabWidth;
    NSPoint buttonOrigin = NSMakePoint(xCoord,0);
    // TODO: take care of tabs that run off the end... right now we're the same as Firebird, except I think we
    // hit our tab limit before we'll run off the end of the average user's windows given the current MIN_TAB_WIDTH
    [tabButtonCell setFrame:NSMakeRect(buttonOrigin.x,buttonOrigin.y,buttonSize.width,buttonSize.height)];
    xCoord += (int)(TAB_GAP+tabWidth);
  }
  
  [self setNeedsDisplay:YES];
}
@end

#pragma mark -
@implementation NSImage (BrowserTabBarViewAdditions)
//an addition to NSImage used by both BrowserTabBarView and TabButtonCell

//tiles an image in the specified rect... even though it should work vertically, these images
//will only look right tiled horizontally.
- (void)paintTiledInRect:(NSRect)rect
{
  NSSize imageSize = [self size];
  NSRect currTile = NSMakeRect( rect.origin.x, rect.origin.y, imageSize.width, imageSize.height );
  
  while( currTile.origin.y < NSMaxY(rect) ) {
    while( currTile.origin.x < NSMaxX(rect) ) {
      // clip the tile as needed
      if( NSMaxX(currTile) > NSMaxX(rect) ) {
        currTile.size.width -= NSMaxX(currTile) - NSMaxX(rect);
      }
      if( NSMaxY(currTile) > NSMaxY(rect) ) {
        currTile.size.height -= NSMaxY(currTile) - NSMaxY(rect);
      }
      NSRect imageRect = NSMakeRect(0, 0, currTile.size.width, currTile.size.height);
      [self compositeToPoint:currTile.origin fromRect:imageRect operation:NSCompositeSourceOver];
      currTile.origin.x += currTile.size.width;
    }
    currTile.origin.y += currTile.size.height;
  }
}
@end